# --- 80 characters ----------------------------------------------------------
# Created by: Laurie 2018/08/12

"""Task runner."""

import json
import uuid
import time
import socket
import threading
import traceback
import logging as lg

from botocore import exceptions as bc_exc

from . import _util
from . import _state_error

_logger = lg.getLogger(__name__)
_host_name = socket.getfqdn(socket.gethostname())


class _TaskExecution:  # TODO: unit-test
    """Execute a task, providing heartbeats and catching failures.

    Args:
        activity (Activity): activity to execute task of
        task_token (str): task token for execution identification
        task_input: task input
        session (_util.AWSSession): session to communicate to AWS with
    """

    def __init__(self, activity, task_token, task_input, *, session=None):
        self.activity = activity
        self.task_token = task_token
        self.task_input = task_input
        self.session = session or _util.AWSSession()
        self._heartbeat_thread = threading.Thread(target=self._heartbeat)
        self._request_stop = False

    def __str__(self):
        return "Execution '%s' of '%s'" % (self.task_token, self.activity)

    def __repr__(self):
        return "%s(%s, %s, len(task_input)=%s, session=%s)" % (
            type(self).__name__,
            repr(self.activity),
            repr(self.task_token),
            len(self.task_input),
            repr(self.session))

    def _send_heartbeat(self):
        """Send a heartbeat."""
        _logger.debug("Sending heartbeat for '%s'" % self)

        try:
            _ = self.session.sfn.send_task_heartbeat(taskToken=self.task_token)
        except bc_exc.ClientError as e:
            if e.response["Error"]["Code"] != "TaskTimedOut":
                raise
            _logger.error("Task execution '%s' timed-out" % self)
            self._request_stop = True

    def _heartbeat(self):
        """Run heartbeat sending."""
        heartbeat = self.activity.heartbeat
        heartbeat = min(max(heartbeat - 5.0, 1.0), heartbeat)
        while True:
            t = time.time()
            if self._request_stop:
                break
            self._send_heartbeat()
            time.sleep(heartbeat - (time.time() - t))

    def _send_failure(self, exc):
        """Report failure."""
        if self._request_stop:
            _logger.warning("Skipping sending failure as we're quitting")
            return

        _logger.info("Report task failure for '%s'" % self, exc_info=exc)

        self._request_stop = True
        cause = traceback.format_exception(type(exc), exc, exc.__traceback__)
        cause = "".join(cause)
        self.session.sfn.send_task_failure(
            taskToken=self.task_token,
            error=type(exc).__name__,
            cause=cause)

    def _send_success(self, res):
        """Report success."""
        if self._request_stop:
            _logger.warning("Skipping sending failure as we're quitting")
            return

        _s = "Report task success for '%s' with output: %s"
        _logger.info(_s % (self, res))

        self._request_stop = True
        self.session.sfn.send_task_success(
            taskToken=self.task_token,
            output=json.dumps(res))

    def run(self):
        """Run task."""
        self._heartbeat_thread.start()
        t = time.time()

        try:
            res = self.activity.call_with(self.task_input)
        except KeyboardInterrupt:
            self.report_cancelled()
            return
        except Exception as e:
            self._send_failure(e)
            return

        _s = "Task '%s' completed in %.6f seconds"
        _logger.debug(_s % (self, time.time() - t))
        self._send_success(res)

    def report_cancelled(self):
        """Cancel a task execution before beginning."""
        if self._request_stop:
            _logger.warning("Skipping sending cancellation as we're quitting")
            return

        _s = "Reporting task failure for '%s' due to cancellation"
        _logger.info(_s % self)

        self._request_stop = True
        self.session.sfn.send_task_failure(
            taskToken=self.task_token,
            error=_state_error.WorkerCancel.__name__,
            cause=str(_state_error.WorkerCancel()))


class Worker:  # TODO: unit-test
    """Worker to poll for task executions.

    Args:
        activity (Activity): activity to poll and run executions of
        name (str): name of worker, used for identification, default: a
            combination of UUID and host's FQDN
        session (_util.Session): session to use for AWS communication
    """

    _task_execution_class = _TaskExecution

    def __init__(self, activity, name=None, *, session=None):
        self.activity = activity
        self.name = name or "%s-%s" % (_host_name, uuid.uuid4())
        self.session = session or _util.AWSSession()

        self._poller = threading.Thread(target=self._worker)
        self._request_finish = False
        self._exc = None

    def __str__(self):
        _s = "%s '%s' on '%s'"
        return _s % (type(self).__name__, self.name, self.activity)

    def __repr__(self):
        return "%s(%s, %s, session=%s)" % (
            type(self).__name__,
            repr(self.activity),
            repr(self.name),
            repr(self.session))

    def _exectute_on(self, task_input, task_token):
        """Execute the provided task.

        Arguments:
            task_input: activity task execution input
            task_token (str): task execuion identifier
        """

        _logger.debug("Got task input: %s" % task_input)

        execution = self._task_execution_class(
            self.activity,
            task_token,
            task_input,
            session=self.session)
        if self._request_finish:
            execution.report_cancelled()
        else:
            execution.run()

    def _poll_and_execute(self):
        """Poll for tasks to execute, then execute any found."""
        while True:
            if self._request_finish:
                break
            _s = "Polling for activity '%s' executions"
            _logger.debug(_s % self.activity)
            resp = self.session.sfn.get_activity_task(
                activityArn=self.activity.arn,
                workerName=self.name)
            if resp.get("taskToken", None) is not None:
                self._exectute_on(json.loads(resp["input"]), resp["taskToken"])

    def _worker(self):
        """Run polling, catching exceptins."""
        _util.assert_valid_name(self.name)
        try:
            self._poll_and_execute()
        except (Exception, KeyboardInterrupt) as e:
            self._exc = e  # send exception to main thread
            self.end()

    def start(self):
        """Start polling."""
        self._poller.start()

    def join(self):
        """Block until polling exit."""
        try:
            self._poller.join()
        except KeyboardInterrupt:
            _logger.info("Quitting polling due to KeyboardInterrupt")
            self.end()
        except Exception:
            self.end()
            raise
        if self._exc is not None:
            raise self._exc

    def end(self):
        """End polling."""
        _logger.info("Worker '%s': waiting on final poll to finish" % self)
        self._request_finish = True

    def run(self):
        """Run worker to poll for and execute specified tasks."""
        self.start()
        self.join()



# --- 80 characters -----------------------------------------------------------
# Created by: Laurie 2018/08/11

"""Package utilities."""

import boto3

import logging as lg
import functools as ft

from . import __name__ as package_name

_logger = lg.getLogger(__name__)

MAX_NAME_LENGTH = 79
INVALID_NAME_CHARACTERS = " \n\t<>{}[]?*$%\\^|~`$,;:/"


def setup_logging():
    """Setup logging for ``sfini``, if logs would otherwise be ignored."""
    package_logger = lg.getLogger(package_name)
    if not package_logger.propagate or package_logger.disabled:
        return
    if not package_logger.hasHandlers():
        handler = lg.StreamHandler()
        fmt = "%(asctime)s \t%(levelname)s \t%(name)s \t%(message)s"
        formatter = lg.Formatter(fmt)
        handler.setFormatter(formatter)
        package_logger.addHandler(handler)
    if not package_logger.isEnabledFor(lg.INFO):
        package_logger.setLevel(lg.INFO)


def cached_property(fn):  # TODO: unit-test
    """Decorate a method as a cached property.

    The wrapped method's result is stored in the instance's ``__cache__``
    dictionary, with the method's name as key.

    Args:
        fn (callable): method to decorate

    Returns:
        property: cached property
    """

    name = fn.__name__

    @ft.wraps(fn)
    def wrapped(self):
        if not hasattr(self, "__cache__"):
            self.__cache__ = {}
        if name not in self.__cache__:
            self.__cache__[name] = fn(self)
        return self.__cache__[name]

    return property(wrapped)


def assert_valid_name(name):  # TODO: unit-test
    """Ensure a valid name of activity, state-machine or state.

    Args:
        name (str): name to analyse

    Raises:
        ValueError: name is invalid
    """

    if len(name) > MAX_NAME_LENGTH:
        raise ValueError("Name is too long: '%s'" % name)
    if any(c in name for c in INVALID_NAME_CHARACTERS):
        raise ValueError("Name contains invalid characters: '%s'" % name)


def collect_paginated(fn, kwargs=None):  # TODO: unit-test
    """Call SFN API paginated endpoint.

    Arguments:
        fn (callable): SFN API function
        kwargs (dict): arguments to ``fn``, default=``{}``

    Returns:
        combined results of paginated API calls
    """

    if kwargs is None:
        kwargs = {}

    if "nextToken" in kwargs:
        raise ValueError("Can't start pagination with 'nextToken'")

    result = fn(**kwargs)
    while "nextToken" in result:
        next_token = result.pop("nextToken")
        next_result = fn(nextToken=next_token, **kwargs)
        for key, value in next_result.items():
            if key == "nextToken":
                result["nextToken"] = value
            else:
                result[key].extend(value)
    return result


class AWSSession:  # TODO: unit-test
    """AWS session, for preconfigure communication with AWS.

    Args:
        session (boto3.Session): session to use
    """

    def __init__(self, session=None):
        self.session = session or boto3.Session()

    def __str__(self):
        _k = self.session.get_credentials().access_key
        return "Session[access key: %s, region: %s]" % (_k, self.region)

    def __repr__(self):
        return "%s(%s)" % (type(self).__name__, repr(self.session))

    @cached_property
    def sfn(self):
        """Step Functions client."""
        setup_logging()
        return self.session.client("stepfunctions")

    @cached_property
    def region(self) -> str:
        """Session AWS region."""
        return self.session.region_name

    @cached_property
    def account_id(self) -> str:
        """Session's account's account ID."""
        _sts = self.session.client("sts")
        return _sts.get_caller_identity()["Account"]



# --- 80 characters -----------------------------------------------------------
# Created by: Laurie 2018/08/11

"""SFN state-machine execution."""

import time
import json
import logging as lg

from . import _util

_logger = lg.getLogger(__name__)


class Execution:  # TODO: unit-test
    """A state-machine execution.

    Args:
        name (str): name of execution
        state_machine (StateMachine): state-machine to execute
        execution_input: execution input (must be JSON-serialisable)
        session (AWSSession): AWS session to use for AWS communication
    """

    def __init__(self, name, state_machine, execution_input, *, session=None):
        self.name = name
        self.state_machine = state_machine
        self.execution_input = execution_input
        self.session = session or _util.AWSSession()

        self._start_time = None
        self._arn = None
        self._output = None

    def __str__(self):
        _s = "%s '%s' on '%s'"
        return _s % (type(self).__name__, self.name, self.state_machine)

    def __repr__(self):
        return "%s(%s, %s, len(execution_input)=%s, session=%s)" % (
            type(self).__name__,
            repr(self.name),
            repr(self.state_machine),
            len(self.execution_input),
            repr(self.session))

    @classmethod
    def from_arn(cls, arn, *, session=None):
        """Construct an ``Execution`` from an existing execution.

        Args:
            arn (str): existing execution ARN
            session (AWSSession): AWS session to use for AWS communication

        Returns:
            Execution: described execution. Note that the ``state_machine``
                attribute will be the ARN of the state-machine, not a
                ``StateMachine`` instance
        """

        session = session or _util.AWSSession()
        resp = session.sfn.describe_execution(executionArn=arn)
        sm_arn = resp["stateMachineArn"]
        self = cls(resp["name"], sm_arn, resp["input"], session=session)
        self._start_time = resp["startDate"]
        self._arn = arn
        self._output = resp.get("output", None)
        return self

    def start(self):
        """Start this state-machine execution."""
        if self._start_time is not None:
            _s = "Execution already started at %s"
            raise RuntimeError(_s % self._start_time)
        _util.assert_valid_name(self.name)
        resp = self.session.sfn.start_execution(
            stateMachineArn=self.state_machine.arn,
            name=self.name,
            input=self.execution_input)
            # input=json.dumps(self.execution_input))
        self._arn = resp["executionArn"]
        self._start_time = resp["startDate"]

    def _get_execution_status(self):
        """Request status of this execution.

        Returns:
            str: execution status
        """

        if self._output is not None:
            return "SUCCEEDED"
        if self._start_time is None:
            raise RuntimeError("Execution not yet started")
        resp = self.session.sfn.describe_execution(executionArn=self._arn)
        if resp["status"] == "SUCCEEDED" and self._output is None:
            self._output = resp["output"]
        return resp["status"]

    @property
    def output(self):
        """Output of execution."""
        if self._output is None:
            if self._get_execution_status() != "SUCCEEDED":
                raise RuntimeError("Execution not yet finished")
        return self._output

    def wait(self, raise_on_error=True):
        """Wait for execution to finish.

        Args:
            raise_on_error (bool): raise error when execution fails
        """

        while True:
            status = self._get_execution_status()
            if status != "RUNNING":
                if status == "SUCCEEDED" or not raise_on_error:
                    break
                raise RuntimeError("Execution %s" % status)
            time.sleep(3.0)

    def stop(self, error_code=None, details=None):
        """Stop an existing execution.

        Args:
            error_code (str): stop reason identification
            details (str): stop reason
        """

        status = self._get_execution_status()
        if status != "RUNNING":
            raise RuntimeError("Cannot stop execution; execution %s" % status)
        _kw = {}
        if error_code:
            _kw["error"] = error_code
        if details:
            _kw["cause"] = details
        resp = self.session.sfn.stop_execution(executionArn=self._arn, **_kw)
        _logger.info("Execution stopped on %s" % resp["stopDate"])

    def print_history(self):
        """Print the execution history."""
        if self._start_time is None:
            raise RuntimeError("Execution not yet started")
        resp = _util.collect_paginated(
            self.session.sfn.get_execution_history,
            kwargs={"executionArn": self._arn})
        for event in resp["events"]:
            id_ = event["id"]
            ts = event["timestamp"]
            t = event["type"]
            prev_id = event["previousEventId"]
            detes = event[event["type"]]
            print("  [%s] %s %s (from %s):\n%s" % (id_, ts, t, prev_id, detes))



# --- 80 characters -----------------------------------------------------------
# Created by: Laurie 2018/08/02

"""SFN state execution error utitilies."""

import logging as lg

_logger = lg.getLogger(__name__)


class WorkerCancel(KeyboardInterrupt):  # TODO: unit-test
    """Workflow execution interrupted by user."""
    def __init__(self, *args, **kwargs):
        _msg = (
            "Activity execution cancelled by user. "
            "This could be due to a `KeyboardInterrupt` during execution, "
            "or the worker was killed during task polling.")
        super().__init__(_msg, *args, **kwargs)


class _ExceptionCondition:  # TODO: unit-test
    @staticmethod
    def _process_exc(exc):
        """Process exception condition.

        Args:
            exc (type or str): exception type-name

        Returns:
            type or str: process exception

        Raises:
            ValueError: bad type-name
        """

        errs = ("*", "ALL", "Timeout", "TaskFailed", "Permissions")

        if isinstance(exc, str):
            if exc not in errs:
                _s = "Error name was '%s', must be one of: %s"
                raise ValueError(_s % (exc, errs))
            return "ALL" if exc == "*" else exc
        elif issubclass(exc, Exception):
            return exc
        else:
            raise TypeError("Error must be exception or predefined string")

    @staticmethod
    def _rules_similar(rule_a, rule_b):
        """Check if rules are similar.

        Args:
            rule_a: LHS rule
            rule_b: RHS rule

        Returns:
            bool: rules are similar
        """

        raise NotImplementedError

    def _collapse_conditions(self, rules):
        """Combine exception rules into lists of exceptions.

        Puts "States.ALL" at the end separately, if it exists.

        Args:
            rules (dict[type or str]): exception rules to group

        Returns:
            list[dict]: rule groups, names put in ``excs``
        """

        all_rule = None
        rule_groups = []
        for exc, rule in rules.items():
            if exc == "ALL":
                all_rule = rule
                continue
            for rule_group in rule_groups:
                if self._rules_similar(rule, rule_group["rule"]):
                    rule_group["excs"].append(exc)
                    break
            else:
                rule_groups.append({"excs": [exc], "rule": rule})
        if all_rule is not None:
            rule_groups.append({"excs": ["ALL"], "rule": all_rule})
        return rule_groups

    def _excs_to_errors(self, excs):
        """Convert exceptions to error codes.

        Arguments:
            excs (list[type or str]): exception conditions

        Returns:
            list[str]: corresponding error codes
        """

        errors = []
        for exc in excs:
            if isinstance(exc, str):
                _ = self._process_exc(exc)
                error = "States." + exc
            elif issubclass(exc, Exception):
                error = str(exc)
            else:
                raise TypeError("Error must be exception or accepted string")
            errors.append(error)
        return errors

    @staticmethod
    def _rule_defn(rule):
        """Get extra definition details from a rule.

        Args:
            rule: exception rule

        Returns:
            dict: extra definition details
        """

        raise NotImplementedError

    def _rule_defns(self, conditions):
        """Build exception rule definitions.

        Collapses rules by similarity into groups, converts exception
        conditions to error codes, then gets the definition for each rule.

        Args:
            conditions (dict[str]): conditions produce definitions from

        Returns:
            list[dict]: definitions
        """

        cond_groups = self._collapse_conditions(conditions)

        defns = []
        for cond_group in cond_groups:
            errors = self._excs_to_errors(cond_group["excs"])
            defn = {"ErrorEquals": errors}
            defn.update(self._rule_defn(cond_group["rule"]))
            defns.append(defn)

        return defns



# --- 80 characters -----------------------------------------------------------
# Created by: Laurie 2018/08/11

"""SFN states."""

import math
import datetime
import logging as lg

from . import _state_error

_logger = lg.getLogger(__name__)


class State:  # TODO: unit-test
    """Abstract state.

    Args:
        name (str): name of state
        comment (str): state description
        state_machine (StateMachine): state-machine this state is a part of
    """

    def __init__(self, name, comment=None, *, state_machine):
        self.name = name
        self.comment = comment
        self.state_machine = state_machine

    def __str__(self):
        return "%s '%s' [%s]" % (
            type(self).__name__,
            self.name,
            self.state_machine.name)

    def __repr__(self):
        return "%s(%s%s%s)" % (
            type(self).__name__,
            repr(self.name),
            "" if self.comment is None else (", " + repr(self.comment)),
            ", state_machine=" + repr(self.state_machine))

    def to_dict(self):
        """Convert this state to a definition dictionary.

        Returns:
            dict: definition
        """

        defn = {"Type": type(self).__name__}
        if self.comment is not None:
            defn["Comment"] = self.comment
        return defn


class _HasNext(State):  # TODO: unit-test
    def __init__(self, name, comment=None, *, state_machine):
        super().__init__(name, comment=comment, state_machine=state_machine)
        self.next = None

    def goes_to(self, state):
        """Set next state after this state finishes.

        Args:
            state (State): state to execute next
        """

        if state not in self.state_machine.states.values():
            _s = "State '%s' is not part of this state-machine"
            raise ValueError(_s % state)

        self.next = state

    def remove_next(self):
        """Remove next state, making this state terminal."""
        self.next = None

    def to_dict(self):
        defn = super().to_dict()
        if self.next is None:
            defn["End"] = True
        else:
            defn["Next"] = self.next.name
        return defn


class _CanRetry(_state_error._ExceptionCondition, State):  # TODO: unit-test
    def __init__(self, name, comment=None, *, state_machine):
        super().__init__(name, comment=comment, state_machine=state_machine)
        self.retries = {}

    def retry_for(
            self,
            exc,
            interval=None,
            max_attempts=None,
            backoff_rate=None):
        """Add a retry condition.

        Args:
            exc (type or str): error for retry to be executed. If a string,
                must be one of '*', 'ALL', 'Timeout', 'TaskFailed', or
                'Permissions' (see AWS Step Functions documentation)
            interval (int): (initial) retry interval (seconds)
            max_attempts (int): maximum number of attempts before re-raising
                error
            backoff_rate (float): retry interval increase factor between
                attempts
        """

        exc = self._process_exc(exc)

        if exc in self.retries:
            raise ValueError("Error '%s' already registered" % exc)

        retry = {}
        if interval is not None:
            retry["interval"] = interval
        if max_attempts is not None:
            retry["max_attempts"] = max_attempts
        if backoff_rate is not None:
            retry["backoff_rate"] = backoff_rate
        self.retries[exc] = retry

    @staticmethod
    def _rules_similar(rule_a, rule_b):
        if rule_a.keys() != rule_b.keys():
            return False
        for k in ("interval", "max_attempts"):
            if k in rule_a and rule_a[k] != rule_b[k]:
                return False
        k = "backoff_rate"
        if k in rule_a and not math.isclose(rule_a[k], rule_b[k]):
            return False
        return True

    @staticmethod
    def _rule_defn(rule):
        defn = {}
        if "interval" in rule:
            defn["IntervalSeconds"] = rule["interval"]
        if "max_attempts" in rule:
            defn["MaxAttempts"] = rule["max_attempts"]
        if "backoff_rate" in rule:
            defn["BackoffRate"] = rule["backoff_rate"]
        return defn

    def to_dict(self):
        defn = super().to_dict()
        retry = self._rule_defns(self.retries)
        if retry:
            defn["Retry"] = retry
        return defn


class _CanCatch(_state_error._ExceptionCondition, State):  # TODO: unit-test
    def __init__(self, name, comment=None, *, state_machine):
        super().__init__(name, comment=comment, state_machine=state_machine)
        self.catches = {}

    def catch(self, exc, next_state):
        """Add a catch clause.

        Args:
            exc (type or str): error for catch clause to be executed. If a
                string, must be one of '*', 'ALL', 'Timeout', 'TaskFailed', or
                'Permissions' (see AWS Step Functions documentation)
            next_state (State): state to execute for catch clause
        """

        if next_state not in self.state_machine.states.values():
            _s = "State '%s' is not part of this state-machine '%s'"
            raise ValueError(_s % (next_state, self.state_machine))

        exc = self._process_exc(exc)

        if exc in self.catches:
            raise ValueError("Error '%s' already registered" % exc)

        self.catches[exc] = next_state

    @staticmethod
    def _rules_similar(next_state_a, next_stateb_b):
        return next_state_a.name == next_stateb_b.name

    @staticmethod
    def _rule_defn(next_state):
        return {"Next": next_state.name}

    def to_dict(self):
        defn = super().to_dict()
        catch = self._rule_defns(self.catches)
        if catch:
            defn["Catch"] = catch
        return defn


class Succeed(State):  # TODO: unit-test
    """End execution successfully.

    Args:
        name (str): name of state
        comment (str): state description
    """

    def __init__(self, name, comment=None, *, state_machine):
        super().__init__(name, comment=comment, state_machine=state_machine)


class Fail(State):  # TODO: unit-test
    """End execution unsuccessfully.

    Args:
        name (str): name of state
        comment (str): state description
        cause (str): failure description
        error (str): name of failure error
    """

    def __init__(
            self,
            name,
            comment=None,
            cause=None,
            error=None,
            *,
            state_machine):
        super().__init__(name, comment=comment, state_machine=state_machine)
        self.cause = cause
        self.error = error

    def to_dict(self):
        defn = super().to_dict()
        defn["Cause"] = self.cause
        defn["Error"] = self.error
        return defn

    def __repr__(self):
        return "%s(%s%s%s%s%s)" % (
            type(self).__name__,
            repr(self.name),
            "" if self.comment is None else (", " + repr(self.comment)),
            "" if self.cause is None else (", " + repr(self.cause)),
            "" if self.error is None else (", " + repr(self.error)),
            ", " + repr(self.state_machine))


class Pass(_HasNext, State):  # TODO: unit-test
    """No-op state, possibly introducing data.

    The name specifies the location of any introduced data.

    Args:
        name (str): name of state
        comment (str): state description
        result: return value of state, stored in the variable ``name``

    Attributes:
        next (State): next state to execute
    """

    def __init__(self, name, comment=None, result=None, *, state_machine):
        super().__init__(name, comment=comment, state_machine=state_machine)
        self.result = result

    def __repr__(self):
        return "%s(%s%s%s%s)" % (
            type(self).__name__,
            repr(self.name),
            "" if self.comment is None else (", " + repr(self.comment)),
            "" if self.result is None else (", " + repr(self.result)),
            ", " + repr(self.state_machine))

    def to_dict(self):
        defn = super().to_dict()
        if self.result is not None:
            defn["ResultPath"] = "$.%s" % self.name
            defn["Result"] = self.result
        return defn


class Wait(_HasNext, State):  # TODO: unit-test
    """Wait for a time before continuing.

    Args:
        name (str): name of state
        until (int or datetime.datetime or str): time to wait. If ``int``, then
            seconds to wait; if ``datetime.datetime``, then time to wait until;
            if ``str``, then name of variable containing seconds to wait for
        comment (str): state description

    Attributes:
        next (State): next state to execute
    """

    def __init__(self, name, until, comment=None, *, state_machine):
        super().__init__(name, comment=comment, state_machine=state_machine)
        self.until = until

    def __repr__(self):
        return "%s(%s%s%s%s)" % (
            type(self).__name__,
            repr(self.name),
            repr(self.until),
            "" if self.comment is None else (", " + repr(self.comment)),
            ", " + repr(self.state_machine))

    def to_dict(self):
        defn = super().to_dict()
        t = self.until
        if isinstance(t, int):
            defn["Seconds"] = t
        elif isinstance(t, datetime.datetime):
            if t.tzinfo is None or t.tzinfo.utcoffset(t) is None:
                raise ValueError("Wait time must be aware")
            defn["Timestamp"] = t.isoformat("T")
        elif isinstance(t, str):
            defn["SecondsPath"] = "$.%s" % t
        else:
            _s = "Invalid type for wait time: %s"
            raise TypeError(_s % type(t).__name__)
        return defn


class Parallel(_HasNext, _CanRetry, _CanCatch, State):  # TODO: unit-test
    """Run states-machines in parallel.

    Args:
        name (str): name of state
        comment (str): state description

    Attributes:
        state_machines (list[StateMachine]): state-machines to run in parallel.
            These state-machines do not need to be registered with AWS Step
            Functions.
        next (State): next state to execute
        retries (dict[Exception or str]): retry conditions
        catches (dict[Exception or str]): handled state errors
    """

    def __init__(self, name, comment=None, *, state_machine):
        super().__init__(name, comment=comment, state_machine=state_machine)
        self.state_machines = []

    def to_dict(self):
        defn = super().to_dict()
        defn["Branches"] = [sm.to_dict() for sm in self.state_machines]
        defn["ResultPath"] = "$.%s" % self.name
        return defn

    def add(self, state_machine):
        """Add a state-machine to be executed.

        The input to the state-machine execution is the input into this
        parallel state. The output of the parallel state is a list of each
        state-machine's output (in order of adding).

        Args:
            state_machine (sfini.StateMachine): state-machine to add. It will
                be run when this task is executed. State-machines do not need
                to be registered with AWS Step Functions
        """

        self.state_machines.append(state_machine)


class Choice(State):  # TODO: unit-test
    """Branch execution based on comparisons.

    Args:
        name (str): name of state
        comment (str): state description

    Attributes:
        choices (list[_ChoiceRule]): choice rules determining branch conditions
        default (State): fall-back state if all comparisons fail
    """

    def __init__(self, name, comment=None, *, state_machine):
        super().__init__(name, comment=comment, state_machine=state_machine)
        self.choices = []
        self.default = None

    def to_dict(self):
        if not self.choices and self.default is None:
            raise RuntimeError("Choice '%s' has no next path")
        defn = super().to_dict()
        defn["Choices"] = [cr.to_dict() for cr in self.choices]
        if self.default is not None:
            defn["Default"] = self.default.name
        return defn

    def add(self, rule):
        """Add a branch.

        Args:
            rule (_ChoiceOp): branch execution condition and specification to
                add

        Raises:
            RuntimeError: rule will go to a state not part of this
                state-machine
        """

        if rule.next_state.state_machine is not self.state_machine:
            _s = "Rule '%s' has next-state which is not part of this "
            _s += "state-machine"
            raise RuntimeError(_s % rule)
        self.choices.append(rule)

    def remove(self, rule):
        """Remove a branch.

        Args:
            rule (_ChoiceOp): branch execution condition and specification to
                remove

        Raises:
            ValueError: if rule is not a registered branch
        """

        if rule not in self.choices:
            raise ValueError("Rule '%s' is not registered with this state")
        self.choices.remove(rule)

    def set_default(self, state):
        """Set the default state to execute when no conditions were met.

        Args:
            state (State): default state to execute
        """

        if self.default is not None:
            _s = "Overwriting current default state '%s'"
            _logger.warning(_s % self.default)
        self.default = state


class Task(_HasNext, _CanRetry, _CanCatch, State):  # TODO: unit-test
    """Activity execution.

    Args:
        name (str): name of state
        activity (Activity or str): activity to execute. If ``Activity``,
            the task is executed by an activity runner. If ``str``, the
            task is run by the AWS Lambda function named ``activity``
        comment (str): state description
        timeout (int): seconds before task time-out

    Attributes:
        next (State): next state to execute
        retries (dict[Exception or str]): retry conditions
        catches (dict[Exception or str]): handled state errors
    """

    def __init__(
            self,
            name,
            activity,
            comment=None,
            timeout=None,
            *,
            state_machine):
        super().__init__(name, comment=comment, state_machine=state_machine)
        self.activity = activity
        self.timeout = timeout

    def __repr__(self):
        to_cm_str = ", timeout=" if self.comment is None else ", "
        _to = "" if self.timeout is None else (to_cm_str + repr(self.timeout))
        return "%s(%s%s%s%s%s)" % (
            type(self).__name__,
            repr(self.name),
            repr(self.activity),
            "" if self.comment is None else (", " + repr(self.comment)),
            _to,
            ", state_machine=" + repr(self.state_machine))

    @property
    def is_lambda(self) -> bool:
        """This task is run by an AWS Lambda Function."""
        return isinstance(self.activity, str)

    def _get_resource_arn(self):
        """Get the activity resource ARN.

        Returns:
            str: activity ARN
        """

        if self.is_lambda:
            region = self.state_machine.session.region
            account = self.state_machine.session.account_id
            _s = "arn:aws:lambda:%s:%s:function:%s"
            return _s % (region, account, self.activity)
        else:
            return self.activity.arn

    def to_dict(self):
        defn = super().to_dict()
        defn["Resource"] = self._get_resource_arn()
        defn["ResultPath"] = "$._results.%s" % self.name
        if self.timeout is not None:
            defn["TimeoutSeconds"] = self.timeout
        if not self.is_lambda:
            defn["HeartbeatSeconds"] = self.activity.heartbeat
        return defn



# --- 80 characters -----------------------------------------------------------
# Created by: Laurie 2018/08/11

"""SFN state-machine."""

import json
import uuid
import datetime
import logging as lg

from . import _util
from . import _execution
from . import _states

_logger = lg.getLogger(__name__)


class StateMachine:  # TODO: unit-test
    """State machine structure for AWS Step Functions.

    Args:
        name (str): name of state-machine
        role_arn (str): AWS ARN for state-machine IAM role
        comment (str): description of state-maching
        timeout (int): execution time-out (seconds)
        session (_util.AWSSession): session to use for AWS communication
    """

    _execution_class = _execution.Execution

    def __init__(
            self,
            name,
            role_arn,
            comment=None,
            timeout=None,
            *,
            session=None):
        self.name = name
        self.role_arn = role_arn
        self.comment = comment
        self.timeout = timeout
        self.session = session or _util.AWSSession()
        self._start_state = None
        self._task_runner_threads = []
        self.states = {}

    def __str__(self):
        return "State-machine '%s' (%s states)" % (self.name, len(self.states))

    def __repr__(self):
        tottl = ", %s" % ("timeout=" if self.comment is None else "")
        return "%s(%s, %s%s%s%s)" % (
            type(self).__name__,
            repr(self.name),
            repr(self.role_arn),
            "" if self.comment is None else ", " + repr(self.comment),
            "" if self.timeout is None else tottl + repr(self.timeout),
            ", session=" + repr(self.session))

    @_util.cached_property
    def arn(self):
        """State-machine generated ARN."""
        region = self.session.region
        account = self.session.account_id
        _s = "arn:aws:states:%s:%s:stateMachine:%s"
        return _s % (region, account, self.name)

    def succeed(self, name, comment=None):
        """Create a succeed state.

        Ends execution successfully.

        Args:
            name (str): name of state
            comment (str): state description

        Returns:
            _states.Succeed: succeed state
        """

        if name in self.states:
            raise ValueError("State name '%s' already registered" % name)
        state = _states.Succeed(name, comment=comment, state_machine=self)
        self.states[name] = state
        return state

    def fail(self, name, comment=None, cause=None, error=None):
        """Create a fail state.

        Ends execution unsuccessfully.

        Args:
            name (str): name of state
            comment (str): state description
            cause (str): failure description
            error (str): name of failure error

        Returns:
            _states.Fail: fail state
        """

        if name in self.states:
            raise ValueError("State name '%s' already registered" % name)
        state = _states.Fail(
            name,
            comment=comment,
            cause=cause,
            error=error,
            state_machine=self)
        self.states[name] = state
        return state

    def pass_(self, name, comment=None, result=None):
        """Create a pass state.

        No-op state, but can introduce data. The name specifies the location of
            the introduced data.

        Args:
            name (str): name of state
            comment (str): state description
            result: return value of state, stored in the variable ``name``

        Returns:
            _states.Pass: pass state
        """

        if name in self.states:
            raise ValueError("State name '%s' already registered" % name)
        state = _states.Pass(
            name,
            comment=comment,
            result=result,
            state_machine=self)
        self.states[name] = state
        return state

    def wait(self, name, until, comment=None):
        """Create a wait state.

        Waits until a time before continuing.

        Args:
            name (str): name of state
            until (int or datetime.datetime or str): time to wait. If ``int``,
                then seconds to wait; if ``datetime.datetime``, then time to
                wait until; if ``str``, then name of variable containing
                seconds to wait for
            comment (str): state description

        Returns:
            _states.Wait: wait state
        """

        if name in self.states:
            raise ValueError("State name '%s' already registered" % name)
        state = _states.Wait(name, until, comment=comment, state_machine=self)
        self.states[name] = state
        return state

    def parallel(self, name, comment=None):
        """Create a parallel state.

        Runs states-machines in parallel. These state-machines do not need to
        be registered with AWS Step Functions.

        The input to each state-machine execution is the input into this
        parallel state. The output of the parallel state is a list of each
        state-machine's output (in order of adding).

        Args:
            name (str): name of state
            comment (str): state description

        Returns:
            _states.Parallel: parallel state
        """

        if name in self.states:
            raise ValueError("State name '%s' already registered" % name)
        state = _states.Parallel(name, comment=comment, state_machine=self)
        self.states[name] = state
        return state

    def choice(self, name, comment=None):
        """Create a choice state.

        Creates branches of possible execution based on conditions.

        Args:
            name (str): name of state
            comment (str): state description

        Returns:
            _states.Choice: choice state
        """

        if name in self.states:
            raise ValueError("State name '%s' already registered" % name)
        state = _states.Choice(name, comment=comment, state_machine=self)
        self.states[name] = state
        return state

    def task(self, name, activity, comment=None, timeout=None):
        """Create a task state.

        Executes an activity.

        Args:
            name (str): name of state
            activity (Activity or str): activity to execute. If ``Activity``,
                the task is executed by an activity runner. If ``str``, the
                task is run by the AWS Lambda function named ``activity``
            comment (str): state description
            timeout (int): seconds before task time-out

        Returns:
            _states.Task: task state
        """

        if name in self.states:
            raise ValueError("State name '%s' already registered" % name)
        state = _states.Task(
            name,
            activity,
            comment=comment,
            timeout=timeout,
            state_machine=self)
        self.states[name] = state
        return state

    def start_at(self, state):
        """Define starting state.

        Args:
            state (State): initial state
        """

        if state.state_machine is not self:
            _s = "State '%s' is not part of this state-machine"
            raise ValueError(_s % state)
        if self._start_state is not None:
            _s = "Overriding start state %s with %s"
            _logger.warning(_s % (self._start_state, state))
        self._start_state = state

    def to_dict(self):
        """Convert this state-machine to a definition dictionary.

        Returns:
            dict: definition
        """

        state_defns = {n: s.to_dict() for n, s in self.states.items()}
        defn = {"StartAt": self._start_state.name, "States": state_defns}
        if self.comment is not None:
            defn["Comment"] = self.comment
        if self.timeout is not None:
            defn["TimeoutSeconds"] = self.timeout
        return defn

    def is_registered(self):
        """See if this state-machine is registered.

        Returns:
            bool: if this state-machine is registered
        """

        resp = _util.collect_paginated(self.session.sfn.list_state_machines)
        arns = {sm["stateMachineArn"] for sm in resp["stateMachines"]}
        return self.arn in arns

    def _sfn_create(self):
        """Create this state-machine in SFN."""
        _logger.info("Creating state-machine '%s' on SFN" % self)
        resp = self.session.sfn.create_state_machine(
            name=self.name,
            definition=json.dumps(self.to_dict(), indent=4),
            roleArn=self.role_arn)
        assert resp["stateMachineArn"] == self.arn
        return resp

    def _sfn_update(self):
        """Update this state-machine in SFN."""
        _logger.info("Updating state-machine '%s' on SFN" % self)
        resp = self.session.sfn.update_state_machine(
            definition=json.dumps(self.to_dict(), indent=4),
            roleArn=self.role_arn,
            stateMachineArn=self.arn)
        return resp

    def register(self, allow_update=False):
        """Register state-machine with AWS Step Functions.

        Args:
            allow_update (bool): if ``True``, allow overwriting of an
                existing state-machine with the same name
        """

        _util.assert_valid_name(self.name)

        if allow_update and self.is_registered():
            resp = self._sfn_update()
            _s = "State machine '%s' updated at %s"
            _logger.info(_s % (self, resp["updateDate"]))
        else:
            resp = self._sfn_create()
            _arn = resp["stateMachineArn"]
            _date = resp["creationDate"]
            _s = "State machine '%s' created with ARN '%s' at %s"
            _logger.info(_s % (self, _arn, _date))

    def deregister(self):
        """Remove state-machine from AWS SFN."""
        if not self.is_registered():
            raise RuntimeError("Cannot de-register unregistered state-machine")

        _logger.info("Deleting state-machine '%s' from SFN" % self)

        _ = self.session.sfn.delete_state_machine(stateMachineArn=self.arn)
        _logger.info("State-machine '%s' de-registered" % self)

    def start_execution(self, execution_input):
        """Start an execution.

        Args:
            execution_input (dict): input to first state in state-machine

        Returns:
            Execution: started execution
        """

        _s = "Starting execution of '%s' with: %s"
        _logger.info(_s % (self, execution_input))

        _now = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
        name = self.name + "_" + _now + "_" + str(uuid.uuid4())[:8]
        execution = self._execution_class(
            name,
            self,
            execution_input,
            session=self.session)
        execution.start()
        return execution

    def list_executions(self, status=None):
        """List all executions of this state-machine.

        Arguments:
            status (str): only list executions with this status. Choose from
                'RUNNING', 'SUCCEEDED', 'FAILED', 'TIMED_OUT' or 'ABORTED'
        """

        kwargs = {"stateMachineArn": self.arn}
        if status is not None:
            kwargs["statusFilter"] = status
        fn = self.session.sfn.list_executions
        resp = _util.collect_paginated(fn, kwargs=kwargs)

        executions = []
        for exec_info in resp["executions"]:
            assert exec_info["stateMachineArn"] == self.arn
            execution = _execution.Execution(
                name=exec_info["name"],
                state_machine=self,
                execution_input=None,
                session=self)
            execution._arn = exec_info["executionArn"]
            execution._start_time = exec_info["startDate"]
            executions.append(execution)

            _s = "Found execution '%s' with status '%s' and stop-date: %s"
            _logger.debug(_s % (execution, exec_info["status"], exec_info["stopDate"]))

        return executions




# --- 100 characters ------------------------------------------------------------------------------
# Created by: Laurie 2018/08/06

"""SFN service helper.

Use in your ``__main__`` module to provide a CLI to your service.
"""

import json
import argparse

from . import _util
from . import _worker


class CLI:
    """``sfini`` command-line interface.

    Arguments:
        state_machine (StateMachine): state-machine to run
        activities (Activities): activities to run
    """

    def __init__(self, state_machine, activities):
        self.state_machine = state_machine
        self.activities = activities

    def _build_parser(self):
        description = self.state_machine.comment or None
        parser = argparse.ArgumentParser(
            description=description)
        parser.add_argument(
            "-V",
            "--version",
            action="version",
            version=self.activities.version)
        parser.add_argument(
            "-v",
            "--verbose",
            action="store_true",
            help="use verbose output")
        subparsers = parser.add_subparsers(
            metavar="COMMAND",
            help="bla",
            dest="command")

        register_parser = subparsers.add_parser(
            "register",
            help="register (or update) activities and state-machine",
            description="register (or update) activities and state-machine")
        register_parser.add_argument(
            "-o",
            "--state-machine-only",
            action="store_true",
            help="only register (or update) state-machine")

        start_parser = subparsers.add_parser(
            "start",
            help="start state-machine execution")
        start_parser.add_argument(
            "input_json",
            help="execution input JSON, use '-' for STDIN")
        start_parser.add_argument(
            "-w",
            "--wait",
            action="store_true",
            help="wait for execution to finish, and print output")

        worker_parser = subparsers.add_parser(
            "worker",
            help="run an acitivity worker",
            description="run an activity worker")
        worker_parser.add_argument(
            "activity_name",
            nargs="*",
            help="name of activity to run")

        executions_parser = subparsers.add_parser(
            "executions",
            help="list executions",
            description="list executions")
        executions_parser.add_argument(
            "-s",
            "--status",
            default=None,
            help="only list executions with this status")

        return parser

    def _execute(self, args, parser):
        if args.command == "register":
            self.state_machine.register(allow_update=args.state_machine_only)
            if not args.state_machine_only:
                self.activities.register()
        elif args.command == "start":
            if args.input_json == "-":
                execution_input = input()
            else:
                with open(args.input_json, "r") as fl:
                    execution_input = json.load(fl)
            execution = self.state_machine.start_execution(execution_input)
            if args.wait:
                execution.wait()
                print(execution.output)
        elif args.command == "worker":
            workers = []
            for activity_name in args.activity_name:
                activity = self.activities.activities[activity_name]
                worker = _worker.Worker(activity)
                worker.start()
                workers.append(worker)
            try:
                for worker in workers:
                    worker.join()
            except KeyboardInterrupt:
                for worker in workers:
                    worker.end()
                    worker.join()
        elif args.command == "executions":
            execs = self.state_machine.list_executions(status=args.status)
            print("Executions:\n" + "\n".join(map(str, execs)))
        else:
            parser.error("Invalid command: %s" % repr(args.command))

    def parse_args(self):
        """Parse command-line arguments and run CLI."""
        _util.setup_logging()
        parser = self._build_parser()
        args = parser.parse_args()
        self._execute(args, parser)



# --- 80 characters -----------------------------------------------------------
# Created by: Laurie 2018/08/12

"""Activity wrapper."""

import inspect
import logging as lg
import functools as ft

from . import _util

_logger = lg.getLogger(__name__)


class Activity:  # TODO: unit-test
    """Activity execution.

    Note that activity names must be unique (within a region). It's
    recommended to put your code's title and version in the activity name.
    ``Activities`` makes this straight-forward.

    An activity is attached to state-machine tasks, and is called when that
    task is executed. A worker registers itself able to run some
    activities using their names.

    Args:
        name (str): name of activity
        fn (callable): function to run activity
        heartbeat (int): seconds between heartbeat during activity running
        session (_util.Session): session to use for AWS communication
    """

    def __init__(self, name, fn, heartbeat=20, *, session=None):
        self.name = name
        self.fn = fn
        self.heartbeat = heartbeat
        self.session = session or _util.AWSSession()
        self.sig = inspect.Signature.from_callable(fn)

    def __call__(self, *args, **kwargs):
        return self.fn(*args, **kwargs)

    def __str__(self):
        return "%s '%s'" % (type(self).__name__, self.name)

    def __repr__(self):
        return type(self).__name__ + "(%s, %s, session=%s)" % (
            repr(self.name),
            repr(self.fn),
            repr(self.session))

    @classmethod
    def from_callable(cls, fn, name, heartbeat=20, *, session=None):
        """Create an activity from the callable.

        Args:
            fn (callable): function to run activity
            name (str): name of activity
            heartbeat (int): seconds between heartbeat during activity running
            session (_util.Session): session to use for AWS communication
        """

        activity = cls(name, fn, heartbeat=heartbeat, session=session)
        ft.update_wrapper(activity, fn)
        return activity

    @_util.cached_property
    def arn(self) -> str:
        """Activity generated ARN."""
        region = self.session.region
        account = self.session.account_id
        _s = "arn:aws:states:%s:%s:activity:%s"
        return _s % (region, account, self.name)

    def register(self):
        """Register activity with AWS."""
        _util.assert_valid_name(self.name)
        resp = self.session.sfn.create_activity(name=self.name)
        assert resp["activityArn"] == self.arn
        _s = "Activity '%s' registered at %s"
        _logger.info(_s % (self, resp["creationDate"]))

    def _get_input_from(self, task_input):
        """Parse task input for execution input.

        Args:
            task_input (dict): task input

        Returns:
            dict: activity input
        """

        kwargs = {}
        for param_name, param in self.sig.parameters.items():
            val = task_input.get(param_name, param.default)
            if val is param.empty:
                _s = "Required parameter '%s' not in task input"
                raise KeyError(_s % param_name)
            kwargs[param_name] = val
        return kwargs

    def call_with(self, task_input):
        """Call with task-input context.

        Args:
            task_input (dict): task input

        Returns:
            function return-value
        """

        kwargs = self._get_input_from(task_input)
        return self.fn(**kwargs)


class Activities:  # TODO: unit-test
    """Activities registration.

    Provides convenience for grouping activities, generating activity
    names, and bulk-registering activities.

    An activity is attached to state-machine tasks, and is called when that
    task is executed. A worker registers itself able to run some activities
    using their names.

    Args:
        name (str): name of activities group, used in prefix of activity
            names
        version (str): version of activities group, used in prefix of
            activity names
        session (_util.Session): session to use for AWS communication

    Attributes:
        activities (dict[str, Activity]): registered activities

    Example:
        >>> activities = Activities("foo", "1.0")
        >>> @activities.activity("myActivity")
        >>> def fn():
        ...     print("hi")
        >>> print(fn.name)
        foo_1.0_myActivity
    """

    _activity_class = Activity

    def __init__(self, name, version="latest", *, session=None):
        self.name = name
        self.version = version
        self.activities = {}
        self.session = session or _util.AWSSession()

    def __str__(self):
        return "%s '%s' [%s]" % (type(self).__name__, self.name, self.version)

    def __repr__(self):
        return "%s(%s, %s, session=%s)" % (
            type(self).__name__,
            repr(self.name),
            repr(self.version),
            repr(self.session))

    @property
    def all_activities(self):
        """All registered activities."""
        return set(self.activities.values())

    def activity(self, name=None, heartbeat=20):
        """Activity function decorator.

        Args:
            name (str): name of activity, default: function name
            heartbeat (int): seconds between heartbeat during activity running
        """

        pref = "%s-%s-" % (self.name, self.version)

        def wrapper(fn):
            suff = fn.__name__ if name is None else name
            if suff in self.activities:
                raise ValueError("Activity '%s' already registered" % suff)
            activity = Activity.from_callable(
                fn,
                pref + suff,
                heartbeat=heartbeat,
                session=self.session)
            self.activities[suff] = activity
            return activity
        return wrapper

    def register(self):
        """Add registered activities to AWS SFN."""
        for activity in self.activities.values():
            activity.register()

    def deregister(self, version=None):
        """Remove activities in AWS SFN.

        Args:
            version (str): version of activities to remove, default: all other
                versions
        """

        # List activities in SFN
        # Determine available versions
        # Remove all activities with requested version
        raise NotImplementedError  # TODO: implement activity deletion



# --- 80 characters -------------------------------------------------------
# Created by: Laurie 2018/08/11

"""Install ``sfini``."""

import pathlib
import setuptools

_parent = pathlib.Path(__file__).parent
_long_description = _parent.joinpath("README.md").read_text()
_version = _parent.joinpath("VERSION").read_text().strip()
_classifiers = [
    "Environment :: Console",
    "Intended Audience :: Developers",
    "Programming Language :: Python :: 3 :: Only",
    "Natural Language :: English",
    "Operating System :: POSIX :: Linux",
    "Operating System :: Microsoft :: Windows",
    "Operating System :: MacOS",
    "License :: OSI Approved :: MIT License"]

setuptools.setup(
    name="sfini",
    version=_version,
    license="MIT",
    author="Laurie",
    author_email="laurie@sitesee.com.au",
    maintainer="Laurie",
    maintainer_email="laurie@sitesee.com.au",
    description="Create AWS Step Functions easily",
    long_description=_long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/Epic_Wink/aws-sfn-service",
    classifiers=_classifiers,
    keywords="aws sfn service step functions",
    packages=setuptools.find_packages(exclude=["tests/*"]),
    python_requires="~=3.3",
    install_requires=["boto3"],
    extras_require={
        "dev": ["pytest", "pytest-cov", "mock", "moto"]})



# --- 80 characters -----------------------------------------------------------
# Created by: Laurie 2018/08/11

"""SFN choice rule operations."""

import datetime
import logging as lg

_logger = lg.getLogger(__name__)


class _ChoiceOp:  # TODO: unit-test
    pass


class _ChoiceRule(_ChoiceOp):  # TODO: unit-test
    def __init__(self, variable_name, comparison_value, next_state):
        self.variable_name = variable_name
        self.comparison_value = comparison_value
        self.next_state = next_state

    def __str__(self):
        return "'%s' %s %s [%s]" % (
            self.variable_name,
            type(self).__name__,
            self.comparison_value,
            self.next_state)

    def __repr__(self):
        return "%s(%s, %s, %s)" % (
            type(self).__name__,
            repr(self.variable_name),
            repr(self.comparison_value),
            repr(self.next_state))

    def to_dict(self):
        """Convert this rule to a definition dictionary.

        Returns:
            dict: definition
        """

        op_name = type(self).__name__
        if op_name.startswith("_"):
            raise RuntimeError("'%s' is not a valid choice rule" % op_name)
        return {
            "Variable": self.variable_name,
            op_name: self.comparison_value,
            "Next": self.next_state.name}


class BooleanEquals(_ChoiceRule):  # TODO: unit-test
    """Compare boolean variable value.

    Args:
        variable_name (str): name of variable to compare
        comparison_value (bool): value to compare against
        next_state (_state.State): state to execute on success
    """

    def __init__(self, variable_name, comparison_value, next_state):
        super().__init__(variable_name, comparison_value, next_state)
        if not isinstance(comparison_value, bool):
            raise TypeError("Boolean comparison value must be `bool`")


class _NumericRule(_ChoiceRule):  # TODO: unit-test
    """Compare numeric variable value.

    Args:
        variable_name (str): name of variable to compare
        comparison_value (int or float): value to compare against
        next_state (_state.State): state to execute on success
    """

    def __init__(self, variable_name, comparison_value, next_state):
        super().__init__(variable_name, comparison_value, next_state)
        if not isinstance(comparison_value, (int, float)):
            _s = "Numeric comparison value must be `int` or `float`"
            raise TypeError(_s)


class NumericEquals(_NumericRule):
    pass


class NumericGreaterThan(_NumericRule):
    pass


class NumericGreaterThanEquals(_NumericRule):
    pass


class NumericLessThan(_NumericRule):
    pass


class NumericLessThanEquals(_NumericRule):
    pass


class _StringdRule(_ChoiceRule):  # TODO: unit-test
    """Compare string variable value.

    Args:
        variable_name (str): name of variable to compare
        comparison_value (str): value to compare against
        next_state (_state.State): state to execute on success
    """

    def __init__(self, variable_name, comparison_value, next_state):
        super().__init__(variable_name, comparison_value, next_state)
        if not isinstance(comparison_value, str):
            raise TypeError("String comparison value must be `str`")


class StringEquals(_StringdRule):
    pass


class StringGreaterThan(_StringdRule):
    pass


class StringGreaterThanEquals(_StringdRule):
    pass


class StringLessThan(_StringdRule):
    pass


class StringLessThanEquals(_StringdRule):
    pass


class _TimestampdRule(_ChoiceRule):  # TODO: unit-test
    """Compare date/time variable value.

    Args:
        variable_name (str): name of variable to compare
        comparison_value (datetime.datetime): value to compare against
        next_state (_state.State): state to execute on success
    """

    def __init__(self, variable_name, comparison_value, next_state):
        super().__init__(variable_name, comparison_value, next_state)
        if not isinstance(comparison_value, datetime.datetime):
            _s = "Timestamp comparison value must be `datetime.datetime`"
            raise TypeError(_s)

    def to_dict(self):
        op_name = type(self).__name__
        if op_name.startswith("_"):
            raise RuntimeError("'%s' is not a valid choice rule")
        t = self.comparison_value
        if t.tzinfo is None or t.tzinfo.utcoffset(t) is None:
            raise ValueError("Comparison time must be aware")
        return {
            "Variable": self.variable_name,
            op_name: t.isoformat("T"),
            "Next": self.next_state.name}


class TimestampEquals(_TimestampdRule):
    pass


class TimestampGreaterThan(_TimestampdRule):
    pass


class TimestampGreaterThanEquals(_TimestampdRule):
    pass


class TimestampLessThan(_TimestampdRule):
    pass


class TimestampLessThanEquals(_TimestampdRule):
    pass


class _LogicalRule(_ChoiceOp):  # TODO: unit-test
    def __init__(self, choice_rules, next_state):
        self.choice_rules = choice_rules
        self.next_state = next_state

    def __str__(self):
        _t = " %s " % type(self).__name__
        _s = _t.join("(%s)" % r for r in self.choice_rules)
        return _s + " [%s]" % self.next_state

    def __repr__(self):
        return "%s(%s, %s)" % (
            type(self).__name__,
            repr(self.choice_rules),
            repr(self.next_state))

    def _get_choice_rule_defns(self):
        """Build choice rule definitions.

        Returns:
            list[dict]: choice rule definitions
        """

        choice_rule_defns = []
        for choice_rule in self.choice_rules:
            defn = choice_rule.to_dict().copy()
            del defn["Next"]
            choice_rule_defns.append(defn)
        return choice_rule_defns

    def to_dict(self):
        """Convert this rule to a definition dictionary.

        Returns:
            dict: definition
        """

        op_name = type(self).__name__
        if op_name.startswith("_"):
            raise RuntimeError("'%s' is not a valid choice rule")
        choice_rule_defns = self._get_choice_rule_defns()
        return {op_name: choice_rule_defns, "Next": self.next_state.name}


class And(_LogicalRule):
    pass


class Or(_LogicalRule):
    pass


class Not(_LogicalRule):  # TODO: unit-test
    def __init__(self, choice_rule, next_state):
        super().__init__([choice_rule], next_state)

    def __str__(self):
        _f = (type(self).__name__, self.choice_rules[0], self.next_state)
        return "%s %s [%s]" % _f

    def __repr__(self):
        return "%s(%s, %s)" % (
            type(self).__name__,
            repr(self.choice_rules[0]),
            repr(self.next_state))

    def _get_choice_rule_defns(self):
        defn = self.choice_rules[0].to_dict().copy()
        del defn["Next"]
        return defn



# --- 80 characters -----------------------------------------------------------
# Created by: Laurie 2018/08/11

"""AWS Step Functions service."""

import pkg_resources

try:
    __version__ = pkg_resources.get_distribution(__name__).version
except pkg_resources.DistributionNotFound:
    __version__ = None

__all__ = [
    "StateMachine",
    "Worker",
    "AWSSession",
    "Activities",
    "Execution",
    "WorkerCancel",
    "And",
    "Or",
    "Not",
    "BooleanEquals",
    "NumericEquals",
    "NumericGreaterThan",
    "NumericGreaterThanEquals",
    "NumericLessThan",
    "NumericLessThanEquals",
    "StringEquals",
    "StringGreaterThan",
    "StringGreaterThanEquals",
    "StringLessThan",
    "StringLessThanEquals",
    "TimestampEquals",
    "TimestampGreaterThan",
    "TimestampGreaterThanEquals",
    "TimestampLessThan",
    "TimestampLessThanEquals",
    "CLI"]

from ._state_machine import StateMachine
from ._worker import Worker
from ._util import AWSSession
from ._activity import Activities
from ._execution import Execution
from ._state_error import WorkerCancel
from ._choice_ops import And
from ._choice_ops import Or
from ._choice_ops import Not
from ._choice_ops import BooleanEquals
from ._choice_ops import NumericEquals
from ._choice_ops import NumericGreaterThan
from ._choice_ops import NumericGreaterThanEquals
from ._choice_ops import NumericLessThan
from ._choice_ops import NumericLessThanEquals
from ._choice_ops import StringEquals
from ._choice_ops import StringGreaterThan
from ._choice_ops import StringGreaterThanEquals
from ._choice_ops import StringLessThan
from ._choice_ops import StringLessThanEquals
from ._choice_ops import TimestampEquals
from ._choice_ops import TimestampGreaterThan
from ._choice_ops import TimestampGreaterThanEquals
from ._choice_ops import TimestampLessThan
from ._choice_ops import TimestampLessThanEquals
from ._cli import CLI
